@@ -2,16 +2,7 @@
 #include <structmember.h>
 
 /*
-TODO: 
-- Python 3 compatibility?
-- PyPy compatibility
-- contains, count
-- Fix automated memory leak tests (eg. check memory usage, run for a while and check it again)
-- Test to optimize by creating a node pool and reuse deallocated nodes instead of allocating new ones
-- Create optimized function for replacing multiple elements in the same function call
-- Register with Hashable
-
-Persistant/Immutable data structures. Unfortunately I have not been able to come
+Persistent/Immutable data structures. Unfortunately I have not been able to come
 up with an implementation that is 100% immutable due to the ref counts used by
 Python internally so the GIL must still be at work...
 
@@ -39,6 +30,15 @@ typedef struct {
   unsigned int refCount;
 } VNode;
 
+#define NODE_CACHE_MAX_SIZE 1024
+
+typedef struct {
+  unsigned int size;
+  VNode* nodes[NODE_CACHE_MAX_SIZE];
+} vNodeCache;
+
+static vNodeCache nodeCache;
+
 typedef struct {
   PyObject_HEAD
   unsigned int count;   // Perhaps ditch this one in favor of ob_size/Py_SIZE()
@@ -48,18 +48,55 @@ typedef struct {
 } PVector;
 
 static PVector* EMPTY_VECTOR = NULL;
+static PyObject* pmap_factory_fn = NULL;
+static PyObject* assoc_in_fn_name = NULL;
+
+static PyObject* assocInPmap(PyObject* keySequence, Py_ssize_t keySize, PyObject* value) {
+  if(pmap_factory_fn == NULL) {
+    // Lazy import pmap factory to avoid circular import problems
+    pmap_factory_fn = PyObject_GetAttrString(PyImport_ImportModule("pyrsistent"), "pmap");
+  }
+
+  PyObject* emptyMap = PyObject_CallFunctionObjArgs(pmap_factory_fn, NULL);
+  PyObject* newSequence = PySequence_GetSlice(keySequence, 1, keySize);
+  PyObject* newMap = PyObject_CallMethodObjArgs(emptyMap,
+						assoc_in_fn_name,
+						newSequence,
+						value,
+						NULL);
+  Py_DECREF(emptyMap);
+  Py_DECREF(newSequence);
+  return newMap;
+}
 
 // No access to internal members
 static PyMemberDef PVector_members[] = {
 	{NULL}  /* Sentinel */
 };
 
-
 #define debug(...)
 // #define debug printf
 
+static VNode* allocNode(void) {
+  if(nodeCache.size > 0) {
+    nodeCache.size--;
+    return nodeCache.nodes[nodeCache.size];
+  }
+
+  return PyMem_Malloc(sizeof(VNode));
+}
+
+static void freeNode(VNode *node) {
+  if(nodeCache.size < NODE_CACHE_MAX_SIZE) {
+    nodeCache.nodes[nodeCache.size] = node;
+    nodeCache.size++;
+  } else {
+    PyMem_Free(node);
+  }
+}
+
 static VNode* newNode(void) {
-  VNode* result = malloc(sizeof(VNode));
+  VNode* result = allocNode();
   memset(result, 0x0, sizeof(VNode));
   result->refCount = 1;
   debug("newNode() %p\n", result);
@@ -70,7 +107,7 @@ static VNode* copyNode(VNode* source) {
   /* NB: Only to be used for internal nodes, eg. nodes that do not
          hold direct references to python objects but only to other nodes. */
   int i;
-  VNode* result = malloc(sizeof(VNode));
+  VNode* result = allocNode();
   debug("copyNode() %p\n", result);
   memcpy(result->items, source->items, sizeof(source->items));
   
@@ -88,11 +125,6 @@ static PVector* emptyNewPvec(void);
 static PVector* copyPVector(PVector *original);
 static void extend_with_item(PVector *newVec, PyObject *item);
 
-static void freeNode(VNode* node) {
-  free(node);
-}
-
-
 static Py_ssize_t PVector_len(PVector *self) {
   return self->count;
 }
@@ -122,9 +154,6 @@ static VNode* nodeFor(PVector *self, int i){
   return NULL;
 }
 
-/*
- Returns a new reference as specified by the PySequence_GetItem function.
-*/
 static PyObject* _get_item(PVector *self, Py_ssize_t pos) {
   VNode* node = nodeFor((PVector*)self, pos);
 
@@ -135,7 +164,9 @@ static PyObject* _get_item(PVector *self, Py_ssize_t pos) {
   return NULL;
 }
 
-
+/*
+ Returns a new reference as specified by the PySequence_GetItem function.
+*/
 static PyObject* PVector_get_item(PVector *self, Py_ssize_t pos) {
   if (pos < 0) {
     pos += self->count;
@@ -354,6 +385,60 @@ static int PVector_traverse(PVector *o, visitproc visit, void *arg) {
 }
 
 
+static PyObject* PVector_index(PVector *self, PyObject *args) {
+  // A direct rip-off of the tuple version
+  Py_ssize_t i, start=0, stop=self->count;
+  PyObject *value;
+  
+  if (!PyArg_ParseTuple(args, "O|O&O&:index", &value,
+			_PyEval_SliceIndex, &start,
+			_PyEval_SliceIndex, &stop)) {
+    return NULL;
+  }
+  
+  if (start < 0) {
+    start += self->count;
+    if (start < 0) {
+      start = 0;
+    }
+  }
+  
+  if (stop < 0) {
+    stop += self->count;
+      if (stop < 0) {
+	stop = 0;
+      }
+  }
+  
+  for (i = start; i < stop && i < self->count; i++) {
+    int cmp = PyObject_RichCompareBool(_get_item(self, i), value, Py_EQ);
+    if (cmp > 0) {
+      return PyInt_FromSsize_t(i);
+    } else if (cmp < 0) {
+      return NULL;
+    }
+  }
+
+  PyErr_SetString(PyExc_ValueError, "PVector.index(x): x not in vector");
+  return NULL;
+}
+
+static PyObject* PVector_count(PVector *self, PyObject *value) {
+  Py_ssize_t count = 0;
+  Py_ssize_t i;
+
+  for (i = 0; i < self->count; i++) {
+    int cmp = PyObject_RichCompareBool(_get_item(self, i), value, Py_EQ);
+    if (cmp > 0) {
+      count++;
+    } else if (cmp < 0) {
+      return NULL;
+    }
+  }
+
+  return PyInt_FromSsize_t(count);
+}
+
 static void copyInsert(void** dest, void** src, Py_ssize_t pos, void *obj) {
   memcpy(dest, src, BRANCH_FACTOR * sizeof(void*));
   dest[pos] = obj;
@@ -361,6 +446,8 @@ static void copyInsert(void** dest, void** src, Py_ssize_t pos, void *obj) {
 
 static PyObject* PVector_append(PVector *self, PyObject *obj);
 
+static PyObject* PVector_assoc_in(PVector *self, PyObject *obj);
+
 static PyObject* PVector_assoc(PVector *self, PyObject *obj);
 
 static PyObject* PVector_subscript(PVector* self, PyObject* item);
@@ -386,12 +473,21 @@ static PyMappingMethods PVector_mapping_methods = {
     NULL
 };
 
+PyDoc_STRVAR(index_doc,
+"V.index(value, [start, [stop]]) -> integer -- return first index of value.\n"
+"Raises ValueError if the value is not present."
+);
+PyDoc_STRVAR(count_doc,
+"V.count(value) -> integer -- return number of occurrences of value");
 
 static PyMethodDef PVector_methods[] = {
 	{"append",      (PyCFunction)PVector_append, METH_O,       "Appends an element"},
-	{"assoc",       (PyCFunction)PVector_assoc,  METH_VARARGS, "Inserts an element at the specified position"},
+	{"assoc",       (PyCFunction)PVector_assoc, METH_VARARGS, "Inserts an element at the specified position"},
 	{"__getitem__", (PyCFunction)PVector_subscript, METH_O|METH_COEXIST, "Subscript"},
 	{"extend",      (PyCFunction)PVector_extend, METH_O|METH_COEXIST, "Extend"},
+	{"assoc_in",    (PyCFunction)PVector_assoc_in, METH_VARARGS, "Insert an element in a nested structure"},
+	{"index",       (PyCFunction)PVector_index,  METH_VARARGS, index_doc},
+	{"count",       (PyCFunction)PVector_count,  METH_O, count_doc},
 	{NULL}
 };
 
@@ -400,7 +496,7 @@ static PyObject * PVectorIter_iter(PyObject *seq);
 static PyTypeObject PVectorType = {
   PyObject_HEAD_INIT(NULL)
   0,
-  "pyrsistent.PVector",                       /* tp_name        */
+  "pvectorc.PVector",                       /* tp_name        */
   sizeof(PVector),                            /* tp_basicsize   */
   0,		                              /* tp_itemsize    */
   (destructor)PVector_dealloc,                /* tp_dealloc     */
@@ -618,11 +714,13 @@ static PyObject *PVector_subscript(PVector* self, PyObject* item) {
 
    These are some optimizations that could be done to this function,
    these are not considered important enough yet though.
+   - Use the PySequence_Fast ops if the iterable is a list or a tuple (which it
+     whould probably often be)
    - Only copy the original tail if it is not full
    - No need to try to increment ref count in tail for the whole tail
 */
 static PyObject* PVector_extend(PVector *self, PyObject *iterable) {
-    PyObject *it;      /* iter(v) */
+    PyObject *it;
     PyObject *(*iternext)(PyObject *);
 
     it = PyObject_GetIter(iterable);
@@ -652,7 +750,6 @@ static PyObject* PVector_extend(PVector *self, PyObject *iterable) {
 
 /*
  Steals a reference to the object that is appended to the list.
- Haven't really decided if this is the right thing to do yet...
 */
 static PyObject* PVector_append(PVector *self, PyObject *obj) {
   assert (obj != NULL);
@@ -713,17 +810,9 @@ static VNode* doAssoc(VNode* node, unsigned int level, unsigned int position, Py
   }
 }
 
-
-/*
- Steals a reference to the object that is inserted in the vector.
-*/
-static PyObject* PVector_assoc(PVector *self, PyObject *args) {
-  PyObject *argObj = NULL;  /* list of arguments */
-  Py_ssize_t position;
-
-  /* The n parses for size, the O parses for a Python object */
-  if(!PyArg_ParseTuple(args, "nO", &position, &argObj)) {
-    return NULL;
+static PyObject* internalAssoc(PVector *self, Py_ssize_t position, PyObject *argObj) {
+  if(position < 0) {
+    position += self->count;
   }
 
   if((0 <= position) && (position < self->count)) {
@@ -753,9 +842,76 @@ static PyObject* PVector_assoc(PVector *self, PyObject *args) {
   }
 }
 
+static PyObject* PVector_assoc_in(PVector *self, PyObject *args) {
+  PyObject *keySequence = NULL;
+  PyObject *value = NULL;
+
+  if(!PyArg_ParseTuple(args, "OO", &keySequence, &value)) {
+    return NULL;
+  }
+
+  Py_ssize_t keySize = PySequence_Size(keySequence);
+  if(keySize == 0) {
+    Py_INCREF(self);
+    return (PyObject*)self;
+  } else {
+    PyObject *index = PySequence_GetItem(keySequence, 0);
+    Py_DECREF(index);
+    Py_ssize_t keyIndex = PyNumber_AsSsize_t(index, NULL);
+    if (keyIndex == -1 && PyErr_Occurred()) {
+      return NULL;
+    }
+
+    if(keySize == 1) {
+      return internalAssoc(self, keyIndex, value);
+    } else if(keyIndex == self->count) {
+      PyObject* newMap = assocInPmap(keySequence, keySize, value);
+      if(newMap == NULL) {
+      	return NULL;
+      }
+
+      PyObject* newVector = PVector_append(self, newMap);
+      Py_DECREF(newMap);
+      return newVector;
+    } else {
+      PyObject* currentItem = PVector_get_item(self, keyIndex);
+      if(currentItem == NULL) {
+	return NULL;
+      }
+
+      PyObject* newSequence = PySequence_GetSlice(keySequence, 1, keySize);
+      PyObject* newItem = PyObject_CallMethodObjArgs(currentItem, assoc_in_fn_name, newSequence, value, NULL);
+      Py_DECREF(currentItem);
+      Py_DECREF(newSequence);
+      if(newItem == NULL) {
+	return NULL;
+      }
+      
+      PyObject* newVector = internalAssoc(self, keyIndex, newItem);
+      Py_DECREF(newItem);
+      return newVector;
+    }
+  }
+}
+
+/*
+ Steals a reference to the object that is inserted in the vector.
+*/
+static PyObject* PVector_assoc(PVector *self, PyObject *args) {
+  PyObject *argObj = NULL;  /* argument to insert */
+  Py_ssize_t position;
+
+  /* The n parses for size, the O parses for a Python object */
+  if(!PyArg_ParseTuple(args, "nO", &position, &argObj)) {
+    return NULL;
+  }
+
+  return internalAssoc(self, position, argObj);
+}
+
 
 static PyMethodDef PyrsistentMethods[] = {
-  {"_pvector", pyrsistent_pvec, METH_VARARGS, "Factory method for persistent vectors"},
+  {"pvector", pyrsistent_pvec, METH_VARARGS, "Factory method for persistent vectors"},
   {NULL, NULL, 0, NULL}
 };
 
@@ -766,18 +922,11 @@ PyMODINIT_FUNC initpvectorc(void) {
   PVectorType.tp_init = NULL;
   PVectorType.tp_new = NULL;
 
-  // TODO: Perhaps make it use a proper constructor instead of a factory method
-  //       and name the type pvector so that testing with instanceof or type becomes
-  //       more straight forward?
-
-
   if (PyType_Ready(&PVectorType) < 0)
     return;
 
   // Register with Sequence to appear as a proper sequence to the outside world
-  PyObject* module = PyImport_ImportModule("collections");
-  PyObject* moduleDict = PyModule_GetDict(module);
-  PyObject* c = PyDict_GetItemString(moduleDict, "Sequence");
+  PyObject* c = PyObject_GetAttrString(PyImport_ImportModule("collections"), "Sequence");
 
   if (PyType_Ready(c->ob_type) < 0) {
     printf("Failed to initialize Sequence!\n");
@@ -791,9 +940,9 @@ PyMODINIT_FUNC initpvectorc(void) {
   PyObject_CallMethod(c, "register", "O", &PVectorType);
 
   m = Py_InitModule("pvectorc", PyrsistentMethods);
-
-  if (m == NULL)
+  if (m == NULL) {
     return;
+  }
 
   SHIFT = __builtin_popcount(BIT_MASK);
   
@@ -801,8 +950,12 @@ PyMODINIT_FUNC initpvectorc(void) {
     EMPTY_VECTOR = emptyNewPvec();
   }
 
+  nodeCache.size = 0;
+
   Py_INCREF(&PVectorType);
   PyModule_AddObject(m, "PVector", (PyObject *)&PVectorType);
+
+  assoc_in_fn_name = PyString_FromString("assoc_in");
 }
 
 
@@ -860,7 +1013,6 @@ PyTypeObject PVectorIterType = {
     0,                                          /* tp_members */
 };
 
-
 static PyObject * PVectorIter_iter(PyObject *seq) {
     PVectorIter *it = PyObject_GC_New(PVectorIter, &PVectorIterType);
     if (it == NULL)
@@ -886,8 +1038,9 @@ static int PVectorIter_traverse(PVectorIter *it, visitproc visit, void *arg) {
 static PyObject *PVectorIter_next(PVectorIter *it) {
     assert(it != NULL);
     PVector *seq = it->it_seq;
-    if (seq == NULL)
+    if (seq == NULL) {
         return NULL;
+    }
 
     if (it->it_index < seq->count) {
         PyObject *item = _get_item(seq, it->it_index);
