@@ -20,7 +20,7 @@ def _comparator(f):
     return wrapper
 
 
-class PVector(Sequence, Hashable):
+class PVector(object):
     """
     Do not instantiate directly, instead use the factory functions :py:func:`pvec` and :py:func:`pvector` to
     create an instance.
@@ -54,7 +54,10 @@ class PVector(Sequence, Hashable):
     (1, 99, 3)
     >>>
     """
-    def __init__(self, c, s, r, t):
+    __slots__ = ('_count', '_shift', '_root', '_tail', '_tail_offset')
+
+    def __new__(cls, c, s, r, t):
+        self = super(PVector, cls).__new__(cls)
         self._count = c
         self._shift = s
         self._root = r
@@ -62,13 +65,14 @@ class PVector(Sequence, Hashable):
 
         # Derived attribute stored for performance
         self._tail_offset = self._count - len(self._tail)
+        return self
 
     def __len__(self):
         return self._count
 
     def __getitem__(self, index):
         if isinstance(index, slice):
-            # There are more condintions than the below where it would be OK to
+            # There are more conditions than the below where it would be OK to
             # return ourself, implement those...
             if index.start is None and index.stop is None and index.step is None:
                 return self
@@ -165,6 +169,12 @@ class PVector(Sequence, Hashable):
         """
         Return a new vector with element at position i replaced with val.
         """
+        if not isinstance(i, (int, long)):
+            raise TypeError("'%s' object cannot be interpreted as an index" % type(i).__name__)
+
+        if i < 0:
+            i += self._count
+
         if 0 <= i < self._count:
             if i >= self._tail_offset:
                 new_tail = list(self._tail)
@@ -292,21 +302,41 @@ class PVector(Sequence, Hashable):
         ret.append(self._new_path(level - SHIFT, tail_node))
         return ret
 
+    def assoc_in(self, keys, val):
+        if not keys:
+            return self
+        elif len(keys) == 1:
+            return self.assoc(keys[0], val)
+        elif keys[0] == self._count:
+            return self.append(pmap().assoc_in(keys[1:], val))
+        else:
+            return self.assoc(keys[0], self[keys[0]].assoc_in(keys[1:], val))
+
+    def index(self, value, *args, **kwargs):
+        return self._tolist().index(value, *args, **kwargs)
+
+    def count(self, value):
+        return self._tolist().count(value)
+
+Sequence.register(PVector)
+Hashable.register(PVector)
 
 _EMPTY_VECTOR = PVector(0, SHIFT, [], [])
 
+
 def _pvector(sequence=[]):
     """
     Factory function, returns a new PVector object containing the elements in elements.
     """
     return _EMPTY_VECTOR.extend(sequence)
 
-pvector = _pvector
 
+pvector = _pvector
 try:
-    from pvectorc import _pvector as _pvectorc
-    pvector = _pvectorc
-except:
+    # Use the C extension as underlying implementation if it is available
+    from pvectorc import pvector as pvector_c
+    pvector = pvector_c
+except ImportError:
     pass
 
 
@@ -316,8 +346,9 @@ def v(*elements):
     """
     return pvector(elements)
 
+
 ####################### PMap #####################################
-class PMap(Mapping):
+class PMap(object):
     """
     Do not instantiate directly, instead use the factory functions :py:func:`pmap` and :py:func:`pmapping` to
     create an instance.
@@ -334,7 +365,7 @@ class PMap(Mapping):
 
     >>> m = v(a=1, b=3)
     >>> m2 = m.assoc('c', 3)
-    >>> m3 = m2.without('a')
+    >>> m3 = m2.dissoc('a')
     >>> m
     {'a': 1, 'b': 3}
     >>> m2
@@ -343,6 +374,8 @@ class PMap(Mapping):
     {'c': 3, 'b': 3}
     >>>
     """
+    __slots__ = ('_size', '_buckets')
+
     def __init__(self, size, buckets):
         self._size = size
         self._buckets = buckets
@@ -361,6 +394,19 @@ class PMap(Mapping):
 
         raise KeyError
 
+    def __contains__(self, key):
+        _, bucket = self._get_bucket(key)
+        if bucket:
+            for k, _ in bucket:
+                if k == key:
+                    return True
+
+            return False
+
+        return False
+
+    get = Mapping.get
+
     def __iter__(self):
         return self.iterkeys()
 
@@ -372,7 +418,7 @@ class PMap(Mapping):
             yield k
 
     # These are more efficient implementations compared to the original
-    # methods that is based on the keys iterator and then calls the
+    # methods that are based on the keys iterator and then calls the
     # accessor functions to access the value for the corresponding key
     def itervalues(self):
         for _, v in self.iteritems():
@@ -399,8 +445,13 @@ class PMap(Mapping):
     def __repr__(self):
         return str(dict(self))
 
+    __eq__ = Mapping.__eq__
+    __ne__ = Mapping.__ne__
     __str__ = __repr__
 
+    def __hash__(self):
+        return hash(tuple(self.iteritems()))
+
     def assoc(self, key, val):
         """
         Return a new map with key and val inserted.
@@ -428,7 +479,7 @@ class PMap(Mapping):
         # Skip reallocation check if there was no conflict
         return PMap(self._size + 1, self._buckets.assoc(index, [kv]))
 
-    def without(self, key):
+    def dissoc(self, key):
         """
         Return a new map without the element specified by key.
         """
@@ -443,6 +494,31 @@ class PMap(Mapping):
 
         return self
 
+    def merge(self, *maps):
+        # Optimization opportunities here
+        if not maps:
+            return self
+        elif len(maps) > 1:
+            merge_map = dict(maps[0])
+            for m in maps[1:]:
+                merge_map.update(m)
+        else:
+            merge_map = maps[0]
+
+        result = self
+        for k, v in merge_map.items():
+            result = result.assoc(k, v)
+        
+        return result
+
+    def assoc_in(self, keys, val):
+        if not keys:
+            return self
+        elif len(keys) == 1:
+            return self.assoc(keys[0], val)
+        else:
+            return self.assoc(keys[0], self.get(keys[0], _EMPTY_PMAP).assoc_in(keys[1:], val))
+
     def _reallocate_to_list(self, new_size):
         new_list = new_size * [None]
         for k, v in chain.from_iterable(x for x in self._buckets if x):
@@ -457,6 +533,9 @@ class PMap(Mapping):
     def _reallocate(self, new_size):
         return pvector(self._reallocate_to_list(new_size))
 
+Mapping.register(PMap)
+Hashable.register(PMap)
+
 
 def _turbo_mapping(initial, pre_size):
     size = pre_size or (2 * len(initial)) or 8
@@ -483,12 +562,6 @@ def _turbo_mapping(initial, pre_size):
 
 _EMPTY_PMAP = _turbo_mapping({}, 0)
 
-def m(**kwargs):
-    """
-    Factory function, inserts all key value arguments into the newly created map.
-    """
-    return pmap(kwargs)
-
 
 def pmap(initial={}, pre_size=0):
     """
@@ -504,9 +577,15 @@ def pmap(initial={}, pre_size=0):
     return _turbo_mapping(initial, pre_size)
 
 
+def m(**kwargs):
+    """
+    Factory function, inserts all key value arguments into the newly created map.
+    """
+    return pmap(kwargs)
+
 ##################### Pset ########################
 
-class PSet(Set):
+class PSet(object):
     """
     Do not instantiate directly, instead use the factory function :py:func:`pset` to create an instance.
 
@@ -515,8 +594,8 @@ class PSet(Set):
     Some examples:
 
     >>> s = pset([1, 2, 3, 1])
-    >>> s2 = s.add(4)
-    >>> s3 = s2.without(2)
+    >>> s2 = s.assoc(4)
+    >>> s3 = s2.dissoc(2)
     >>> s
     pset([1, 2, 3])
     >>> s2
@@ -525,6 +604,8 @@ class PSet(Set):
     pset([1, 3, 4])
     >>>
     """
+    __slots__ = ('_map',)
+
     def __init__(self, m):
         self._map = m
 
@@ -542,22 +623,40 @@ class PSet(Set):
 
     __str__ = __repr__
 
+    def __hash__(self):
+        return hash(self._map)
+
     @classmethod
     def _from_iterable(cls, it, pre_size=8):
         return PSet(pmap({k: True for k in it}, pre_size=pre_size))
 
-    def add(self, element):
+    def assoc(self, element):
         return PSet(self._map.assoc(element, True))
 
-    def without(self, element):
-        return PSet(self._map.without(element))
+    def dissoc(self, element):
+        return PSet(self._map.dissoc(element))
 
+    # This is not very beautiful. If we avoid inheriting from PSet we can use the
+    # __slots__ concepts (which requires a new style class) and hopefully save some memory.
+    __le__ = Set.__le__
+    __lt__ = Set.__lt__
+    __gt__ = Set.__gt__
+    __ge__ = Set.__ge__
+    __eq__ = Set.__eq__
+    __ne__ = Set.__ne__
+
+    __and__ = Set.__and__
+    __or__ = Set.__or__
+    __sub__ = Set.__sub__
+    __xor__ = Set.__xor__
+
+    isdisjoint = Set.isdisjoint
+
+Set.register(PSet)
+Hashable.register(PSet)
 
 _EMPTY_PSET = PSet(_EMPTY_PMAP)
 
-def s(*args):
-    return pset(args)
-
 
 def pset(sequence=[], pre_size=8):
     """
@@ -568,3 +667,7 @@ def pset(sequence=[], pre_size=8):
         return _EMPTY_PSET
 
     return PSet._from_iterable(sequence, pre_size=pre_size)
+
+
+def s(*args):
+    return pset(args)
