@@ -6,6 +6,9 @@ def test_literalish_works():
     assert s() is pset()
     assert s(1, 2) == pset([1, 2])
 
+def test_supports_hash():
+    assert hash(s(1, 2)) == hash(s(1, 2))
+
 
 def test_contains_elements_that_it_was_initialized_with():
     initial = [1, 2, 3]
@@ -17,14 +20,14 @@ def test_contains_elements_that_it_was_initialized_with():
 
 def test_is_immutable():
     s1 = pset([1])
-    s2 = s1.add(2)
+    s2 = s1.assoc(2)
 
-    assert set(s1) == set([1])
-    assert set(s2) == set([1, 2])
+    assert s1 == pset([1])
+    assert s2 == pset([1, 2])
 
-    s3 = s2.without(1)
-    assert set(s2) == set([1, 2])
-    assert set(s3) == set([2])
+    s3 = s2.dissoc(1)
+    assert s2 == pset([1, 2])
+    assert s3 == pset([2])
 
 
 def test_is_iterable():
@@ -53,4 +56,14 @@ def test_behaves_set_like():
 def test_str():
     assert str(pset([1, 2, 3])) == "pset([1, 2, 3])"
 
+def test_is_disjoint():
+    s1 = pset([1, 2, 3])
+    s2 = pset([3, 4, 5])
+    s3 = pset([4, 5])
+
+    assert not s1.isdisjoint(s2)
+    assert s1.isdisjoint(s3)
+
+
+
 pytest.main()
